@@ -1,7 +1,7 @@
 import importlib, collections
 
 from enum import Enum
-from cro_validate.enum import DataType
+from cro_validate.enum import DataType, VersionMutationType
 from cro_validate.classes.configuration_classes import Config
 import cro_validate.classes.parameter_classes as Parameters
 import cro_validate.classes.schema_classes as Schemas
@@ -9,6 +9,19 @@ import cro_validate.classes.name_strategy_classes as NameStrategies
 import cro_validate.classes.util_classes as Utils
 
 
+def instantiate(fqn, *args, **kw):
+	module_name, class_name = Utils.ClassName.class_fqn_parts(fqn)
+	if module_name == 'builtins' and class_name == 'NoneType':
+		return None
+	module = importlib.import_module(module_name)
+	_class = getattr(module, class_name)
+	if isinstance(_class, type(Enum)):
+		result = _class[args[0]]
+	else:
+		result = _class(*args, **kw)
+	return result
+
+
 class Meta:
 	def initialize(self, definition, **kw):
 		raise NotImplementedError()
@@ -108,6 +121,8 @@ class Definition:
 				deprecated,
 				internal,
 				rules,
+				version,
+				version_conversion,
 				meta
 			):
 		def _set_member(name, value):
@@ -125,6 +140,8 @@ class Definition:
 		_set_member('deprecated', deprecated)
 		_set_member('internal', internal)
 		_set_member('rules', rules)
+		self.version = version  # ephemeral (managed by Index)
+		self.version_conversion = version_conversion
 		_set_member('meta', meta)
 		# Name
 		######
@@ -194,7 +211,15 @@ class Definition:
 		items = []
 		i = 0
 		for entry in value:
-			item = Index.validate_input(None, field_fqn + '[' + str(i) + ']', field_name, self.data_format, entry, **rules_kw)
+			item = Index.validate_input(
+					validated=None,
+					field_fqn=field_fqn + '[' + str(i) + ']',
+					field_name=field_name,
+					definition_or_name=self.data_format,
+					value=entry,
+					version=self.get_version(),
+					**rules_kw
+				)
 			items.append(item[field_name])
 			i = i + 1
 		results[field_name] = items
@@ -282,6 +307,12 @@ class Definition:
 	def get_name(self):
 		return self.name
 
+	def get_version(self):
+		return self.version
+
+	def get_version_conversion(self):
+		return self.version_conversion
+
 	def get_data_type(self):
 		return self.data_type
 
@@ -339,7 +370,7 @@ class Definition:
 					continue
 				fields[name] = getattr(model, name)
 				if fields[name] is None:
-					fields[name] = Schemas.Field(definition_name=Index.get(name))
+					fields[name] = Schemas.Field(input_name=name, definition_name=Index.get(name))
 				elif fields[name].definition_name is None:
 					fields[name].definition_name = name
 		for k in fields:
@@ -362,6 +393,17 @@ class Definition:
 
 
 class DefinitionJsonDeserializer:
+	def __init__(self, root):
+		self.namespace = {k:k for k in root}
+		for k in root:
+			if 'aliases' not in root[k]:
+				continue
+			for k1 in root[k]['aliases']:
+				if k1 in self.namespace:
+					raise Config.exception_factory.create_internal_error(k1, 'Input definition already exists.')
+				self.namespace[k1] = k
+		self.root = root
+
 	def _get_dict_value(self, idx, k, default_value=Utils.Empty):
 		if k in idx:
 			return idx[k]
@@ -375,16 +417,26 @@ class DefinitionJsonDeserializer:
 			k = 'internal'
 		tgt[k] = v
 
-	def _deserialize_schema_field(self, root, name, obj):
+	def _get_root_obj(self, k):
+		root_k = self.namespace[k]
+		result = self.root[root_k]
+		return result
+
+	def _get_definition_name(self, k):
+		result = self.namespace[k]
+		return result
+
+	def _deserialize_schema_field(self, name, obj):
 		kw = {}
 		if 'default_value' in obj:
 			if obj['default_value_type'] != 'cro_validate.classes.util_classes.Empty':
-				kw['default_value'] = self._instantiate(obj['default_value_type'], obj['default_value'])
+				kw['default_value'] = instantiate(obj['default_value_type'], obj['default_value'])
 		kw['definition_name'] = name
 		if 'definition_name' in obj and obj['definition_name'] is not None:
 			kw['definition_name'] = obj['definition_name']
-		if not Index.exists(kw['definition_name']):
-			dependent_definition_profile = self.deserialize(root, kw['definition_name'])
+		definition_name = self._get_definition_name(kw['definition_name'])
+		if not Index.exists(definition_name):
+			dependent_definition_profile = self.deserialize(definition_name)
 			Index.register_definition(**dependent_definition_profile)
 		self._set(obj, kw, 'ignored')
 		self._set(obj, kw, 'input_name', default_value=name)
@@ -394,7 +446,7 @@ class DefinitionJsonDeserializer:
 		field = Schemas.Field(**kw)
 		return field
 
-	def _deserialize_schema(self, root, obj, name):
+	def _deserialize_schema(self, obj, name):
 		kw = {}
 		self._set(obj, kw, 'allow_unknown_fields')
 		self._set(obj, kw, 'case_sensitive')
@@ -412,19 +464,19 @@ class DefinitionJsonDeserializer:
 				if 'kw' in entry:
 					init_action_kw = entry['kw']
 				init_action_class_name = entry['type']
-				init_action = self._instantiate(init_action_class_name, **init_action_kw)
+				init_action = instantiate(init_action_class_name, **init_action_kw)
 				field_init_actions.append(init_action)
 		if 'dependency_resolver' in obj:
 			resolver_class_name = obj['dependency_resolver']['type']
 			resolver_kw = {}
 			if 'kw' in obj['dependency_resolver']:
 				resolver_kw = obj['dependency_resolver']['kw']
-			resolver = self._instantiate(resolver_class_name, **resolver_kw)
+			resolver = instantiate(resolver_class_name, **resolver_kw)
 			kw['dependency_resolver'] = resolver
 		if 'inherits' in obj:
 			for inherited_definition_name in obj['inherits']:
 				if not Index.exists(inherited_definition_name):
-					dependent_definition_profile = self.deserialize(root, inherited_definition_name)
+					dependent_definition_profile = self.deserialize(inherited_definition_name)
 					Index.register_definition(**dependent_definition_profile)
 				parent_definition = Index.get(inherited_definition_name)
 				inherited_fields = parent_definition.list_fields()
@@ -432,7 +484,7 @@ class DefinitionJsonDeserializer:
 					model[field_name] = inherited_fields[field_name]
 		if 'model' in obj:
 			for field_name in obj['model']:
-				field = self._deserialize_schema_field(root, field_name, obj['model'][field_name])
+				field = self._deserialize_schema_field(field_name, obj['model'][field_name])
 				model[field_name] = field
 		initialized_model = {}
 		for action in field_init_actions:
@@ -449,23 +501,38 @@ class DefinitionJsonDeserializer:
 		schema = Schemas.Schema(**kw)
 		return schema
 
-	def _instantiate(self, fqn, *args, **kw):
-		module_name, class_name = Utils.ClassName.class_fqn_parts(fqn)
-		if module_name == 'builtins' and class_name == 'NoneType':
-			return None
-		module = importlib.import_module(module_name)
-		_class = getattr(module, class_name)
-		if isinstance(_class, type(Enum)):
-			result = _class[args[0]]
-		else:
-			result = _class(*args, **kw)
-		return result
-
-	def _deserialize_rule(self, root, obj):
-		rule = self._instantiate(obj['type'], **obj['config'])
+	def _deserialize_rule(self, obj):
+		rule = instantiate(obj['type'], **obj['config'])
 		return rule
 
-	def deserialize(self, root, name):
+	def _deserialize_version(self, obj):
+		result = {}
+		mutations = []
+		for mutation in obj['mutations']:
+			action = mutation['action']
+			action = VersionMutationType[action]
+			if action == VersionMutationType.AddField:
+				mutations.append(Utils.VersionMutation(action, mutation['config'], None))
+			elif action == VersionMutationType.RemoveField:
+				mutations.append(Utils.VersionMutation(action, mutation['config'], None))
+			elif action == VersionMutationType.MutateField:
+				mutations.append(Utils.VersionMutation(action, mutation['config'], None))
+			else:
+				raise Exception('Not Implemented')
+			if 'definition_name' in mutation['config']:
+				definition_name = mutation['config']['definition_name']
+				if not Index.exists(definition_name):
+					dependent_definition_profile = self.deserialize(definition_name)
+					Index.register_definition(**dependent_definition_profile)
+		result['mutations'] = mutations
+		if 'conversion' in obj:
+			config = obj['conversion']['config']
+			type_fqn = obj['conversion']['type']
+			conversion = instantiate(type_fqn, **config)
+			result['conversion'] = conversion
+		return result
+
+	def deserialize(self, name):
 		kw = {
 			'name': Utils.Empty,
 			'aliases': Utils.Empty,
@@ -480,41 +547,52 @@ class DefinitionJsonDeserializer:
 			'meta': Utils.Empty,
 			'rules': Utils.Empty
 		}
-		namespace = {k:k for k in root}
-		for k in root:
-			if 'aliases' not in root[k]:
-				continue
-			for k1 in root[k]['aliases']:
-				namespace[k1] = k
-		obj = root[namespace[name]]
+		obj = self._get_root_obj(name)
 		# Data Format
 		#############
 		if 'data_type' in obj:
 			kw['data_type'] = DataType[obj['data_type']]
 		if 'data_format' in obj:
 			if obj['data_format_type'] == 'cro_validate.classes.schema_classes.Schema':
-				kw['data_format'] = self._deserialize_schema(root, obj['data_format'], name)
+				kw['data_format'] = self._deserialize_schema(obj['data_format'], name)
 			else:
-				kw['data_format'] = self._instantiate(obj['data_format_type'], obj['data_format'])
+				kw['data_format'] = instantiate(obj['data_format_type'], obj['data_format'])
 		# Default Value
 		###############
 		if 'default_value' in obj:
 			if obj['default_value_type'] == 'builtins.NoneType':
 				kw['default_value'] = None
 			elif obj['default_value_type'] != 'cro_validate.classes.util_classes.Empty':
-				kw['default_value'] = self._instantiate(obj['default_value_type'], obj['default_value'])
+				kw['default_value'] = instantiate(obj['default_value_type'], obj['default_value'])
 		# Rules
 		#######
 		if 'rules' in obj:
 			rules = []
 			for rule in obj['rules']:
-				rule_obj = self._deserialize_rule(obj, rule)
+				rule_obj = self._deserialize_rule(rule)
 				rules.append(rule_obj)
 			kw['rules'] = rules
 		# Aliases
 		#########
 		if 'aliases' in obj:
-			kw['aliases'] = set(obj['aliases'])
+			aliases = obj['aliases']
+			if isinstance(aliases, str):
+				aliases = {aliases}
+			else:
+				aliases = set(aliases)
+			kw['aliases'] = aliases
+		# Versions
+		##########
+		if 'base_version' in obj:
+			kw['base_version'] = obj['base_version']
+		if 'base_version_conversion' in obj:
+			config = obj['base_version_conversion']['config']
+			type_fqn = obj['base_version_conversion']['type']
+			kw['base_version_conversion'] = instantiate(type_fqn, **config)
+		if 'versions' in obj:
+			kw['versions'] = {}
+			for version in obj['versions']:
+				kw['versions'][version] = self._deserialize_version(obj['versions'][version])
 		# Simple
 		########
 		kw['name'] = name
@@ -532,15 +610,58 @@ class Index:
 	_data_type_rule_exceptions = set()
 	_data_format_rules = {}
 	_data_format_rule_exceptions = set()
+	_versions = ['base']
+	_versions_idx = {'base': 0}
 
-	def get(definition_or_name):
-		if isinstance(definition_or_name, Definition):
-			return definition_or_name
-		definition_name = str(definition_or_name)
+	def add_version(version):
+		if version in Index._versions_idx:
+			return
+		latest_version = Index.get_latest_version()
+		for k in Index._idx:
+			latest_def = Index.get(k, version=latest_version)
+			Index._idx[k][version] = latest_def
+		Index._versions_idx[version] = len(Index._versions)
+		Index._versions.append(version)
+
+	def list_versions():
+		result = []
+		result.extend(Index._versions)
+		return result
+
+	def get_latest_version():
+		return Index._versions[-1]
+
+	def version_exists(version):
+		if version in Index._versions_idx:
+			return True
+		return False
+
+	def get(definition_or_name, version=Utils.Empty):
+		definition_name = definition_or_name
+		if isinstance(definition_name, Definition) is True:
+			definition_name = definition_or_name.get_name()
+		if Utils.Empty.isempty(version) is not True:
+			if version not in Index._versions:
+				raise Config.exception_factory.create_internal_error(definition_name, 'Unknown/unregistered version requested: {0}'.format(version))
 		resolved = Config.definition_name_resolver.resolve(Index._idx, definition_name)
 		if resolved is None:
-			raise Config.exception_factory.create_input_error(definition_name, 'Definition name resolution failed (Unknown definition name).')
-		return Index._idx[resolved]
+			raise Config.exception_factory.create_internal_error(definition_name, 'Definition name resolution failed (Unknown definition name).')
+		versions = Index._idx[resolved]
+		# Latest
+		########
+		if Utils.Empty.isempty(version) is True:
+			latest_version = Index.get_latest_version()
+			return versions[latest_version]
+		# Base
+		######
+		if version == 'base':
+			for next_version in Index._versions:
+				if next_version in versions:
+					return versions[next_version]
+			raise Config.exception_factory.create_internal_error(definition_name, 'Base version not found: {0}'.format(version))
+		# Specified
+		###########
+		return versions[version]
 
 	def exists(name):
 		resolved = Config.definition_name_resolver.resolve(Index._idx, name)
@@ -562,13 +683,73 @@ class Index:
 		return result
 
 	def from_json_dict(root):
-		deserializer = DefinitionJsonDeserializer()
+		deserializer = DefinitionJsonDeserializer(root)
 		for k in root:
 			if Index.exists(k):
 				continue
-			profile = deserializer.deserialize(root, k)
+			profile = deserializer.deserialize(k)
 			Index.register_definition(**profile)
 
+	def mutate_model(definition_name, base_model, versions, target_version):
+		# Init Mutation
+		###############
+		mutated_model = {}
+		if isinstance(base_model, dict):
+			for k in base_model:
+				if k.startswith('_'):
+					continue
+				if isinstance(base_model[k], Schemas.Field):
+					mutated_model[k] = base_model[k]
+				elif isinstance(base_model[k], type(None)):
+					mutated_model[k] = Schemas.Field(input_name=k)
+		else:
+			for k in dir(base_model):
+				if k.startswith('_'):
+					continue
+				v = getattr(base_model, k)
+				if isinstance(v, Schemas.Field):
+					mutated_model[k] = v
+				elif isinstance(v, type(None)):
+					mutated_model[k] = Schemas.Field(input_name=k)
+		# Mutate
+		########
+		# Mutate in-order until we reach the target version
+		found = False
+		for version in Index._versions:
+			if version not in versions:
+				continue
+			mutations = versions[version]['mutations']
+			for mutation in mutations:
+				if mutation.action == VersionMutationType.AddField:
+					field = Schemas.Field(**mutation.config)
+					field_name = field.output_name
+					mutated_model[field_name] = field
+				elif mutation.action == VersionMutationType.RemoveField:
+					field = Schemas.Field(**mutation.config)
+					del mutated_model[field.output_name]
+				elif mutation.action == VersionMutationType.MutateField:
+					old_field = mutated_model[mutation.config['field_name']]
+					config = {
+							k:mutation.config[k]
+							for k in mutation.config
+							if k != 'field_name'
+						}
+					new_field = Schemas.Field(
+							input_name=old_field.input_name,
+							output_name=old_field.output_name
+						)
+					new_field.mutate(**config)
+					mutated_model[new_field.output_name] = new_field
+			if version == target_version:
+				found = True
+				break
+		if found is not True:
+			raise Config.exception_factory.create_internal_error(
+					definition_name,
+					'Target version not found: {0}.'.format(target_version)
+				)
+		return mutated_model
+
 	def register_definition(
 				name,
 				aliases,
@@ -581,9 +762,16 @@ class Index:
 				deprecated,
 				internal,
 				rules,
-				meta
+				meta,
+				base_version=Utils.Empty,
+				base_version_conversion=Utils.Empty,
+				versions=Utils.Empty
 			):
-		definition = Definition(
+		if Utils.Empty.isempty(base_version) is True:
+			base_version = 'base'
+		if Utils.Empty.isempty(base_version_conversion) is True:
+			base_version_conversion = None
+		base_definition = Definition(
 				name=name,
 				aliases=aliases,
 				description=description,
@@ -595,11 +783,68 @@ class Index:
 				deprecated=deprecated,
 				internal=internal,
 				rules=rules,
+				version=base_version,
+				version_conversion=base_version_conversion,
 				meta=meta
 			)
-		definition_name = definition.get_name()
+		name = base_definition.get_name()
+		if base_version not in Index._versions_idx:
+			raise Config.exception_factory.create_internal_error(
+					name,
+					'Base version not found: {0}.'.format(base_version)
+				)
+		definition_versions = {}
+		latest_def_version = base_definition
+		if Utils.Empty.isempty(versions) is True:
+			versions = []
+		initial_version_index = Index._versions_idx[base_version]
+		for version in Index._versions[initial_version_index:]:
+			if version not in versions:
+				latest_def_version = Definition(
+						name=latest_def_version.name,
+						aliases=latest_def_version.aliases,
+						description=latest_def_version.description,
+						data_type=latest_def_version.data_type,
+						data_format=latest_def_version.data_format,
+						default_value=latest_def_version.default_value,
+						examples=latest_def_version.examples,
+						nullable=latest_def_version.nullable,
+						deprecated=latest_def_version.deprecated,
+						internal=latest_def_version.internal,
+						rules=latest_def_version.rules,
+						version=version,
+						version_conversion=latest_def_version.version_conversion,
+						meta=latest_def_version.meta
+					)
+				definition_versions[version] = latest_def_version
+				continue
+			mutated_data_format = data_format
+			if data_type == DataType.Object:
+				mutated_data_format = data_format.copy()
+				mutated_data_format.model = Index.mutate_model(name, mutated_data_format.model, versions, version)
+			version_conversion = None
+			if 'conversion' in versions[version]:
+				version_conversion = versions[version]['conversion']
+			mutated_definition = Definition(
+					name=name,
+					aliases=aliases,
+					description=description,
+					data_type=data_type,
+					data_format=mutated_data_format,
+					default_value=default_value,
+					examples=examples,
+					nullable=nullable,
+					deprecated=deprecated,
+					internal=internal,
+					rules=rules,
+					version=version,
+					version_conversion=version_conversion,
+					meta=meta
+				)
+			definition_versions[version] = mutated_definition
+			latest_def_version = mutated_definition
 		names = set()
-		names.add(definition_name)
+		names.add(name)
 		if Utils.Empty.isempty(aliases):
 			pass
 		elif isinstance(aliases, str):
@@ -607,11 +852,15 @@ class Index:
 		else:
 			names.update(aliases)
 		for entry in names:
-			if definition_name in Index._idx:
-				raise Config.exception_factory.create_internal_error(definition_name, 'Input definiton already exists.')
+			if entry not in Index._idx:
+				continue
+			raise Config.exception_factory.create_internal_error(
+					entry,
+					'Input definiton already exists.'
+				)
 		for entry in names:
-			Index._idx[entry] = definition
-		return definition
+			Index._idx[entry] = definition_versions
+		return Index.get(name)
 
 	def register_data_type_rule(data_type, rule, exceptions=[]):
 		if data_type not in Index._data_type_rules:
@@ -664,13 +913,78 @@ class Index:
 				field_name,
 				definition_or_name,
 				value,
+				version,
 				**rules_kw
 			):
-		definition = Index.get(definition_or_name)
+		definition = Index.get(definition_or_name, version)
 		results = Parameters.Index.ensure(validated)
 		definition.validate(results, field_fqn, field_name, definition, value, **rules_kw)
 		return results
 
+
+	def mutate_input_version(
+				validated,
+				field_fqn,
+				field_name,
+				definition_or_name,
+				value,
+				src_version,
+				target_version,
+				skip_validation=False,
+				**rules_kw
+			):
+		src_definition = Index.get(definition_or_name, src_version)
+		target_definition = Index.get(definition_or_name, target_version)
+		results = Parameters.Index.ensure(validated)
+		results[src_definition.get_name()] = value
+		src_index = Index._versions_idx[src_version]
+		target_index = Index._versions_idx[target_version]
+		versions = Index._idx[src_definition.get_name()]
+		if skip_validation is False:
+			Index.validate_input(results, field_fqn, field_name, src_definition, value, src_version, **rules_kw)
+			value = results[src_definition.get_name()]
+		if src_index < target_index:
+			i = src_index
+			last_version = None
+			seen_conversions = set()
+			while i < target_index:
+				next_version = Index._versions[i]
+				next_definition = Index.get(definition_or_name, next_version)
+				i = i + 1
+				if last_version == next_definition.get_version():
+					continue
+				last_version = next_definition.get_version()
+				conversion = next_definition.get_version_conversion()
+				if conversion is None:
+					continue
+				if conversion in seen_conversions:
+					continue
+				value = conversion.increment(field_fqn, next_definition, value)
+				seen_conversions.add(conversion)
+		else:
+			i = src_index
+			last_version = None
+			seen_conversions = set()
+			while i > target_index:
+				prev_version = Index._versions[i]
+				prev_definition = Index.get(definition_or_name, prev_version)
+				i = i - 1
+				if last_version == prev_definition.get_version():
+					continue
+				last_version = prev_definition.get_version()
+				conversion = prev_definition.get_version_conversion()
+				if conversion is None:
+					continue
+				if conversion in seen_conversions:
+					continue
+				value = conversion.decrement(field_fqn, prev_definition, value)
+				seen_conversions.add(conversion)
+		if src_version != target_version:
+			results = {src_definition.get_name(): value}
+			if skip_validation is False:
+				results = Index.validate_input(results, field_fqn, field_name, target_definition, value, target_version, **rules_kw)
+		return results
+
 	def ensure_alias(name, alias):
 		definition = Index.get(name)
 		if alias not in Index._idx:
@@ -689,3 +1003,9 @@ class Index:
 
 	def clear():
 		Index._idx = {}
+		Index._data_type_rules = {}
+		Index._data_type_rule_exceptions = set()
+		Index._data_format_rules = {}
+		Index._data_format_rule_exceptions = set()
+		Index._versions = ['base']
+		Index._versions_idx = {'base': 0}
