@@ -44,6 +44,18 @@ class Schema:
 		if self.model_name is None:
 			self.model_name = model.__class__.__name__
 
+	def copy(self):
+		result = Schema(
+				model=self.model,
+				model_name=self.model_name,
+				allow_unknown_fields=self.allow_unknown_fields,
+				display_name=self.display_name,
+				case_sensitive=self.case_sensitive,
+				load_defaults=self.load_defaults,
+				dependency_resolver=self.dependency_resolver,
+				return_unknown_fields=self.return_unknown_fields
+			)
+		return result
 
 	def to_json_dict(self):
 		def _is_default(name, value):
@@ -92,6 +104,11 @@ class Field:
 	unvalidated = False
 	default_value = Utils.Empty
 
+	def _set_member(self, name, value):
+		if Utils.Empty.isempty(value):
+			return
+		setattr(self, name, value)
+
 	def __init__(
 				self,
 				input_name,
@@ -102,22 +119,32 @@ class Field:
 				unvalidated=Utils.Empty,
 				default_value=Utils.Empty
 			):
-		def _set_member(name, value):
-			if Utils.Empty.isempty(value):
-				return
-			setattr(self, name, value)
-		_set_member('input_name', input_name)
-		_set_member('output_name', output_name)
-		_set_member('definition_name', definition_name)
-		_set_member('required', required)
-		_set_member('ignored', ignored)
-		_set_member('unvalidated', unvalidated)
-		_set_member('default_value', default_value)
+		self._set_member('input_name', input_name)
+		self._set_member('output_name', output_name)
+		self._set_member('definition_name', definition_name)
+		self._set_member('required', required)
+		self._set_member('ignored', ignored)
+		self._set_member('unvalidated', unvalidated)
+		self._set_member('default_value', default_value)
 		if self.input_name is None:
 			raise Exceptions.InternalError('Field', 'Input name must be specified')
 		if self.output_name is None:
 			self.output_name = self.input_name
 
+	def mutate(
+				self,
+				definition_name=Utils.Empty,
+				required=Utils.Empty,
+				ignored=Utils.Empty,
+				unvalidated=Utils.Empty,
+				default_value=Utils.Empty
+			):
+		self._set_member('definition_name', definition_name)
+		self._set_member('required', required)
+		self._set_member('ignored', ignored)
+		self._set_member('unvalidated', unvalidated)
+		self._set_member('default_value', default_value)
+
 	# Broken
 	def to_json_dict(self, field_name):
 		def _is_default(name, value):
@@ -390,7 +417,7 @@ class ModelValidator:
 		result = []
 		if isinstance(vector, dict):
 			for k in vector:
-				if self.schema.case_sensitive:
+				if self.schema.case_sensitive is False:
 					k = k.lower()
 				if k in self.input_value_names:
 					result.append(self.input_value_names[k])
@@ -405,7 +432,7 @@ class ModelValidator:
 						continue
 					if callable(getattr(vector, attribute_name)):
 						continue
-					if self.schema.case_sensitive:
+					if self.schema.case_sensitive is False:
 						attribute_name = attribute_name.lower()
 					if attribute_name in self.input_value_names:
 						result.append(self.input_value_names[attribute_name])
@@ -428,7 +455,9 @@ class ModelValidator:
 				return source[value_name]
 		elif hasattr(source, value_name):
 			return getattr(source, value_name)
-		raise Config.exception_factory.create_input_error(fqn, 'Missing required value.')
+		display_name = self.get_field_input_display_name(value_name)
+		field_fqn = Config.definition_name_strategy.get_fqn(fqn, display_name)
+		raise Config.exception_factory.create_input_error(field_fqn, 'Missing required value.')
 
 	def _get_output_value(self, value_name, source):
 		if isinstance(source, dict):
@@ -482,6 +511,8 @@ class ModelValidator:
 		result = Parameters.Index.ensure(validated)
 		normalized = values
 		if isinstance(values, dict):
+			if self.schema.case_sensitive is False:
+				values = {k.lower():values[k] for k in values}
 			normalized = Parameters.Index(values)
 		field_names = self._list_field_input_names(normalized)
 		kw = Parameters.Index()
@@ -535,6 +566,7 @@ class ModelValidator:
 					field_name=field_name,
 					definition_or_name=field_definition,
 					value=kw[field_name],
+					version=field_definition.version,
 					**rules_kw)
 			for definition_result_entry in definition_result:
 				self._set_output_value(
